/* shell.cpp - Implements a serial shell
 * Copyright (C) 2010 Antoine Mercier-Linteau
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "shell.h"
#include "HardwareSerial.h"
#include "base.h"
#include <string.h>
#include <stdlib.h>

#define SUCCESS 0
#define SPECIFIC_FAILURE 1
#define INVALID_ARGS 2
#define UNEXPECTED_ARGUMENTS 3
#define WRONG_SYNTAX 4

Shell::Shell(ShortcutConverter* sc, USBKeyboard* kb):
		position(0),
		level(0),
		command_char(0),
		shortcut_converter(sc),
		keyboard(kb)
{

}

void Shell::prompt(void)
{
	Serial.print_P(PSTR("base:$ "));
}

void Shell::take_char(char incoming)
{
	if(level >= MAX_COMMAND_BUFFER_SIZE)
	{
		Serial.println("");
		Serial.println_P(PSTR("Command too long!"));
		level = 0;
		Serial.flush();
		free(command_buffer);
		return;
	}

	if(command_char)
	{
		if(command_char == 27)
		{
			command_char = incoming;
			return;
		}

		if(command_char == '[')
		{
			switch(incoming)
			{
				case 'D':
					if(position > 0) //Left arrow
					{
						position--;
						Serial.print((char)8);
					}
					break;
				case 'C': //Right arrow
					if(position < level)
					{
						Serial.print(command_buffer[position++]);
					}
					break;
				case 72: //Home
					for(; position > 0; position--)
					{
						Serial.print((char)8);
					}
					break;
				case 70:
					for(; position < level; position++)
					{
						Serial.print(command_buffer[position]);
					}
				default: break;
			}
		}
		command_char = 0;
		return;
	}

	if(incoming == 27) //Escape
	{
		command_char = 27;
		return;
	}
	else if(incoming == 8) //Backspace
	{
		if(position > 0)
		{
			Serial.print((char)8);
			position--;
			for(uint8_t i = position ; i <= level; i++)
			{
				command_buffer[i] = command_buffer[i + 1];
			}
			level--;
			for(uint8_t i = position ; i < level; i++)
			{
				Serial.print(command_buffer[i]);
			}
			Serial.print(' ');
			for(uint8_t i = position ; i <= level; i++)
			{
				Serial.print((char)8);
			}

		}
		return;
	}
	else if(incoming == 127) //Delete
	{
		if(position < level)
		{
			for(uint8_t i = position ; i <= level; i++)
			{
				command_buffer[i] = command_buffer[i + 1];
			}
			level--;
			for(uint8_t i = position ; i < level; i++)
			{
				Serial.print(command_buffer[i]);
			}
			Serial.print(' ');
			for(uint8_t i = position ; i <= level; i++)
			{
				Serial.print((char)8);
			}

		}
		return;
	}
	else if(incoming < ' ' && incoming > '~' && incoming != 10 && incoming != 13)
	{
		return;
	}

	if(level == 0)
	{
		end = MIN_COMMAND_BUFFER_SIZE - 1;
		command_buffer = (char*)malloc(MIN_COMMAND_BUFFER_SIZE);
		position = 0;
		command_char = 0;
	}

	if(level > end)
	{
		char* old_command_buffer = command_buffer;
		command_buffer = (char*)malloc(end + 1 + STEP_SIZE);;
		memcpy(command_buffer, old_command_buffer, end + 1);
		end += STEP_SIZE;
		free(old_command_buffer);
	}


    if(incoming == ';' || incoming == 10 || incoming == 13)
	{

    	Serial.println(""); //Jumps to next line
    	if(verbosity)
    	{
			Serial.print_P(PSTR("Number of chars received: "));
			Serial.println(level, DEC);
    	}

    	command_buffer[level] = ';';
        parse_command();
		Serial.println("");
		prompt();
		level = 0;
		Serial.flush();
		free(command_buffer);
		return;
	}

    if(position != level)
    {
    	for(uint8_t i = 0 ; i <= level - position; i++)
    	{
    		command_buffer[level - i + 1] = command_buffer[level - i];
    	}
    	for(uint8_t i = position ; i <= level; i++)
    	{
    		Serial.print(command_buffer[i]);
    	}
    	for(uint8_t i = position ; i <= level; i++)
    	{
    		Serial.print((char)8);
    	}
    	level++;
    }
    command_buffer[position] = incoming;
	Serial.print(incoming); //echoes the received character
	if(position == level)
	{
		level++;
	}
	position++;

}

void Shell::parse_command()
{
	unsigned char argc = 0;
	char* argv[10];
	bool is_arg = true;

	for( unsigned char i = 0; i <= level; i++)
	{
		if(command_buffer[i] == ';' || command_buffer[i] == 10 || command_buffer[i] == 13)
		{
			command_buffer[i] = '\0';
			break;
		}
		if(command_buffer[i] == ' ')
		{
			command_buffer[i] = '\0';
			is_arg = true;
		}
		else if(is_arg)
		{
			argv[argc] = command_buffer + i;
			argc++;
			is_arg = false;
		}
	}

	if(verbosity)
	{
		Serial.println("Received command structure: ");
		for(unsigned char i = 0; i < argc; i++)
		{
			Serial.print("   ");
			Serial.print(i, DEC);
			Serial.print(": ");
			Serial.println(argv[i]);
		}
	}

	execute_command(argc, argv);
}

void Shell::execute_command( unsigned char argc, char* argv[])
{
	uint8_t result;

	if(!strcmp(argv[0], "help")){ result = do_help(argc, argv); }
	else if(!strcmp(argv[0], "verbosity")){ result = do_verbosity(argc, argv); }
	else if(!strcmp(argv[0], "info")){ result = do_info(argc, argv); }
	//else if(!strcmp(argv[0], "key")){ result = do_key(argc, argv); } //Not implemented
	else if(!strcmp(argv[0], "map")){ result = do_map(argc, argv); }
	else if(!strcmp(argv[0], "sim_type")){ result = do_sim_type(argc, argv); }
	else if(!strcmp(argv[0], "sim_packet")){ result = do_sim_packet(argc, argv); }
	else if(!strcmp(argv[0], "inspect")){ result = do_inspect(argc, argv); }
	else
	{
		Serial.println_P(PSTR("Unrecognized!"));
		return;
	}

	switch(result)
	{
		case SUCCESS: break; //Do nothing if sucess
		case SPECIFIC_FAILURE: Serial.println_P(PSTR("Command failed")); break;
		case INVALID_ARGS: Serial.println_P(PSTR("Invalid arguments!")); break;
		case UNEXPECTED_ARGUMENTS: Serial.println_P(PSTR("Unexpected arguments!")); break;
		case WRONG_SYNTAX: Serial.println_P(PSTR("Wrong syntax!")); break;
		default: Serial.println_P(PSTR("Unknown error code!")); break;
	}
}

uint8_t Shell::do_help(unsigned char argc, char* argv[])
{
	Serial.println_P(PSTR("This is the help"));
	return SUCCESS;
}

uint8_t Shell::do_verbosity(unsigned char argc, char* argv[])
{
	if(argc == 1)
	{
		Serial.print_P(PSTR("verbosity is "));
		if(verbosity) { Serial.println_P(PSTR("ON"));}
		else { Serial.println("OFF"); }
	}
	else if(!strcmp(argv[1], "on"))
	{
		verbosity = true;
		Serial.println_P(PSTR("verbosity on"));
	}
	else if(!strcmp(argv[1], "off"))
	{
		verbosity = false;
		Serial.println(PSTR("verbosity off"));
	}
	else
	{
		return INVALID_ARGS; //returns error code for INVALID_ARGS
	}
	return SUCCESS;
}

uint8_t Shell::do_info(unsigned char argc, char* argv[])
{
	Serial.print_P(PSTR("Number of packets received: "));
	Serial.println(number_of_packets_received, DEC);
	Serial.print_P(PSTR("Number of shortcuts bound: "));
	Serial.println(shortcut_converter->get_map()->size, DEC);
	Serial.print_P(PSTR("Firmware revision: "));
	Serial.println(FIRMWARE_REVISION, DEC);

	return SUCCESS;
}

uint8_t Shell::do_map(unsigned char argc, char* argv[])
{
	if(argc == 1)
	{
		Serial.print(shortcut_converter->get_map()->size, DEC);
		Serial.println_P(PSTR(" entries in map"));
		Serial.println_P(PSTR("Remote ID | Button ID | Modifiers | Key combination"));

		for(unsigned char i = 0; i < shortcut_converter->get_map()->size; i++)
		{
			if(shortcut_converter->get_map()->entries[i].txid)
			{
				Serial.print(shortcut_converter->get_map()->entries[i].txid, DEC);
			}
			else { Serial.print('*'); }
			Serial.print(" | ");
			if(shortcut_converter->get_map()->entries[i].button_id)
			{
				Serial.print(shortcut_converter->get_map()->entries[i].button_id, DEC);
			}
			else { Serial.print('*'); }
			Serial.print(" | ");

			bool next = false;

			for(uint8_t j = 0; j < 8; j++)
			{
				if( shortcut_converter->get_map()->entries[i].key0 & 1<<j )
				{
					if(next){ Serial.print(" + "); }
					char* name = USBKeyboard::usage_id_to_string(224 + j);
					Serial.print(name);
					next = true;
					free(name);
				}

			}
			Serial.print(" | ");

			next = false;

			if(shortcut_converter->get_map()->entries[i].key1 != 0)
			{
				char* name = USBKeyboard::usage_id_to_string(shortcut_converter->get_map()->entries[i].key1);
				Serial.print(name);
				free(name);
				next = true;
			}
			if(shortcut_converter->get_map()->entries[i].key2 != 0)
			{
				if(next){ Serial.print(" + "); }
				char* name = USBKeyboard::usage_id_to_string(shortcut_converter->get_map()->entries[i].key2);
				Serial.print(name);
				free(name);
				next = true;
			}
			if(shortcut_converter->get_map()->entries[i].key3 != 0)
			{
				if(next){ Serial.print(" + "); }
				char* name = USBKeyboard::usage_id_to_string(shortcut_converter->get_map()->entries[i].key3);
				Serial.print(name);
				free(name);
			}
			Serial.println("");
		}

		return SUCCESS;
	}
	else if(argc == 2)
	{
		if(!strcmp(argv[1], "-r"))
		{
			shortcut_converter->reset();
			Serial.println_P(PSTR("Map has been reset!"));
		}

		return SUCCESS;
	}
	else if(argc >= 4)
	{
		if(!strcmp(argv[1], "-d") && argc == 4)
		{
			if(*argv[2] == '*'){ *argv[2] = 0;}
			if(*argv[3] == '*'){ *argv[3] = 0;}

			switch(shortcut_converter->delete_entry(atoi(argv[2]), atoi(argv[3])))
			{
				case 1: Serial.println_P(PSTR("Entry not found!")); break;
				case 2: Serial.println_P(PSTR("Done")); break;
				default: break;
			}

			return SUCCESS;
		}
		else if(!strcmp(argv[1], "-s") && argc == 4)
		{
			unsigned char lower = atoi(argv[2]);
			unsigned char upper = atoi(argv[3]);

			Serial.println_P(PSTR("ID   Name"));
			Serial.println_P(PSTR("--------------"));

			for(unsigned char i = lower; i<upper; i++)
			{
				Serial.print(i, DEC);
				Serial.print_P(PSTR("    "));
				char* usage_name = USBKeyboard::usage_id_to_string(i);
				if(!usage_name)
				{
					return SPECIFIC_FAILURE;
				}
				else
				{
					Serial.println(usage_name);
					free(usage_name);
				}
			}

			return SUCCESS;
		}
		else if(argc > 3 && argc <= 7)
		{
			unsigned char modifiers = 0;
			unsigned char position = 0;
			unsigned char keys[3] = {0, 0, 0};

			for(unsigned char i = 3; i < argc; i++)
			{
				unsigned char id = USBKeyboard::string_to_usage_id(argv[i]);
				if(!id)
				{
					Serial.println_P(PSTR("Unrecognised key"));
					return SPECIFIC_FAILURE;
				}
				if(id >= 224 && id <= 231)
				{
					if(verbosity)
					{
						Serial.println_P(PSTR("modifier identified"));
					}
					modifiers |= 1 << (id - 224);
				}
				else
				{
					if(verbosity)
					{
						Serial.println_P(PSTR("key identified"));
					}
					keys[position++] = id;
				}
			}

			if(*argv[1] == '*'){ *argv[1] = 0;}
			if(*argv[2] == '*'){ *argv[2] = 0;}

			switch( shortcut_converter->save_entry(
					atoi(argv[1]),
					atoi(argv[2]),
					modifiers,
					keys[0],
					keys[1],
					keys[2]) )
			{
				case -1: Serial.println_P(PSTR("Entry replaced"));break;
				case 0: Serial.println_P(PSTR("Entry added")); break;
				case 1: Serial.println_P(PSTR("EEPROM full")); break;
				case 2: Serial.println_P(PSTR("Map full")); break;
				default: return SPECIFIC_FAILURE;
			}
			return SUCCESS;
		}
		else
		{
			return UNEXPECTED_ARGUMENTS;
		}
	}
	else
	{
		return WRONG_SYNTAX;
	}

}

uint8_t Shell::do_sim_type(unsigned char argc, char* argv[])
{
	if(argc > 1)
	{
		unsigned char position = 1;
		char keys[4] = {0, 0, 0, 0};



		for(uint8_t i = 1; i < argc; i++)
		{
			Serial.print_P(PSTR("Finding key "));
			Serial.print(argv[i]);
			uint8_t id = USBKeyboard::string_to_usage_id(argv[i]);
			if(!id)
			{
				Serial.println_P(PSTR(": Unrecognised key"));
				return SPECIFIC_FAILURE;
			}
			if(id >= 224 && id <= 231)
			{
				Serial.println_P(PSTR(": modifier identified"));
				keys[0] |= 1 << (id - 224);
			}
			else
			{
				if(position > 3)
				{
					Serial.println_P(PSTR("Cannot accept more keys"));
				}
				else
				{
					Serial.println_P(PSTR(": key identified"));
					keys[position++] = id;
				}
			}
		}

		keyboard->take_report(keys);

	}
	else
	{
		return UNEXPECTED_ARGUMENTS;
	}
	return SUCCESS;
}
uint8_t Shell::do_sim_packet(unsigned char argc, char* argv[])
{
	if(argc == 3)
	{
		shortcut_converter->take(atoi(argv[1]), atoi(argv[2]));
	}
	else
	{
		return UNEXPECTED_ARGUMENTS;
	}
	return SUCCESS;
}

uint8_t Shell::do_inspect(unsigned char argc, char* argv[])
{
	if(argc == 2)
	{
		Serial.println(ShortcutConverter::inspect(atoi(argv[1])), DEC);
	}
	else
	{
		return UNEXPECTED_ARGUMENTS;
	}
	return SUCCESS;
}
