/* shortcut_converter.cpp - emulates a keyboard
 * Copyright (C) 2010 Antoine Mercier-Linteau
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "shortcut_converter.h"
#include <avr/eeprom.h>
#include <stdlib.h>
#include "base.h"
#include "HardwareSerial.h"

#define EEPROM_MAP_ADDR 4
#define EEPROM_SIZE 1024

ShortcutConverter::ShortcutConverter(void):
	start(0),
	end(0)
{
	load_map();
}

uint8_t ShortcutConverter::available(void)
{
	if(end >= start){ return end - start; }
	else { return SHORTCUT_BUFFER_SIZE - start + end; }
}

int8_t ShortcutConverter::take(uint8_t txid, uint8_t button_id)
{
	if(verbosity)
	{
		Serial.print("Received shortcut ");
		Serial.print(txid, DEC);
		Serial.print(" ");
		Serial.println(button_id, DEC);
	}
	uint8_t ret = 0;
	uint8_t id;

	id = find_entry(txid, button_id);

	if(id != ERROR_ID && ((end + 1) % SHORTCUT_BUFFER_SIZE != start))
	{
		shortcut_buffer[end] = id;

		end = (end + 1) % SHORTCUT_BUFFER_SIZE;
		ret = 0;
	}
	else { ret = 1; }

	id = find_entry(0, button_id);

	if(id != ERROR_ID && ((end + 1) % SHORTCUT_BUFFER_SIZE != start))
	{
		shortcut_buffer[end] = id;

		end = (end + 1) % SHORTCUT_BUFFER_SIZE;
		ret = 0;
	}
	else { ret = 1; }

	id = find_entry(txid, 0);

	if(id != ERROR_ID && ((end + 1) % SHORTCUT_BUFFER_SIZE != start))
	{
		shortcut_buffer[end] = id;

		end = (end + 1) % SHORTCUT_BUFFER_SIZE;
		ret = 0;
	}
	else { ret = 1; }

	id = find_entry(0, 0);

	if(id != ERROR_ID && ((end + 1) % SHORTCUT_BUFFER_SIZE != start))
	{
		shortcut_buffer[end] = id;

		end = (end + 1) % SHORTCUT_BUFFER_SIZE;
		ret = 0;
	}
	else { ret = 1; }

	return ret;
}

int8_t ShortcutConverter::take(uint8_t buffer[], uint8_t length)
{
	if(length < 3){ return 1; }
	for(uint8_t i = 2; i < buffer[0]; i++)
	{
		if(!take(buffer[1], buffer[i]))
		{
			return 1;
		}
	}
	return 0;
}

uint8_t ShortcutConverter::read(char dest[])
{
	uint8_t id = shortcut_buffer[start];

	start = (start + 1) % SHORTCUT_BUFFER_SIZE;

	dest[0] = map.entries[id].key0;
	dest[1] = map.entries[id].key1;
	dest[2] = map.entries[id].key2;
	dest[3] = map.entries[id].key3;

	return 0;
}



void ShortcutConverter::load_map(void)
{
	if(map.entries)
	{
		free(map.entries);
	}

	map.size = (uint8_t)eeprom_read_byte((uint8_t*)EEPROM_MAP_ADDR);

	map.entries = (entry*)malloc(sizeof(entry)*map.size);

	if(map.entries)
	{
		eeprom_read_block(map.entries, (uint8_t*)(EEPROM_MAP_ADDR + 1), map.size*sizeof(entry));
	}

}

shortcut_map* const  ShortcutConverter::get_map(void)
{
	return &map;
}

int8_t ShortcutConverter::save_entry( uint8_t txid,
	uint8_t button_id,
	uint8_t key0,
	uint8_t key1,
	uint8_t key2,
	uint8_t key3)
{
	uint8_t i = find_entry(txid, button_id);

	if(i != ERROR_ID)
	{
		if(verbosity)
		{
			Serial.print("Replacing entry at index ");
			Serial.println(i, DEC);
		}
		map.entries[(uint8_t)i].key0 = key0;
		map.entries[(uint8_t)i].key1 = key1;
		map.entries[(uint8_t)i].key2 = key2;
		map.entries[(uint8_t)i].key3 = key3;
		eeprom_write_block( map.entries + i, (uint8_t*)(EEPROM_MAP_ADDR + 1 + i*sizeof(entry)), sizeof(entry));
		return -1;
	}

	if(map.size == 0xFE)
	{
		return 2; //Maximum number of keys reached.
	}

	if(EEPROM_MAP_ADDR + 1 + (map.size + 1) * sizeof(entry) > EEPROM_SIZE)
	{
		return 1; //EEPROM is full
	}

	if(verbosity)
	{
		Serial.print("Adding entry at index ");
		Serial.println(map.size, DEC);
	}
	map.size++;
	eeprom_write_byte((uint8_t*)EEPROM_MAP_ADDR, map.size);

	entry new_entry = {txid, button_id, key0, key1, key2, key3};

	if(verbosity)
	{
		Serial.print("Added as ");
		Serial.print(new_entry.txid, DEC);
		Serial.print(" ");
		Serial.print(new_entry.button_id, DEC);
		Serial.print(" ");
		Serial.print(new_entry.key0, BIN);
		Serial.print(" ");
		Serial.print(new_entry.key1, DEC);
		Serial.print(" ");
		Serial.print(new_entry.key2, DEC);
		Serial.print(" ");
		Serial.println(new_entry.key3, DEC);
	}

	eeprom_write_block(&new_entry, (uint8_t*)(EEPROM_MAP_ADDR + 1 + (map.size - 1)*sizeof(entry)), sizeof(entry));

	load_map();

	return 0;
}

int8_t ShortcutConverter::delete_entry(uint8_t txid, uint8_t button_id)
{
	uint8_t i = find_entry(txid, button_id);

	if(i == ERROR_ID)
	{
		return 1; //entry not found
	}

	if(map.size > 1 && i != (map.size - 1))
	{
		eeprom_write_block(&map.entries[map.size-1], (uint8_t*)(EEPROM_MAP_ADDR + 1 + i*sizeof(entry)), sizeof(entry));
	}

	eeprom_write_byte((uint8_t*)EEPROM_MAP_ADDR, map.size - 1);

	load_map();

	return 0;
}

uint8_t ShortcutConverter::find_entry(uint8_t txid, uint8_t button_id)
{
	if(verbosity)
	{
		Serial.print("Map entry for txid ");
		Serial.print(txid, DEC);
		Serial.print(" and button_id ");
		Serial.print(button_id, DEC);
	}

	for(char i = 0; i< map.size; i++)
	{
		if(map.entries[i].txid == txid && map.entries[i].button_id == button_id)
		{
			if(verbosity)
			{
				Serial.print(" found at ");
				Serial.println(i, DEC);
			}
			return i;
		}
	}
	if(verbosity)
	{
		Serial.println(" not found");
	}
	return ERROR_ID;
}

void ShortcutConverter::reset(void)
{
	eeprom_write_byte((uint8_t*)EEPROM_MAP_ADDR, 0);
	load_map();
}

uint8_t ShortcutConverter::inspect(int addr)
{
	return eeprom_read_byte((uint8_t*)addr);
}



