/* kbhid.cpp - implements a HID compatible USB keyboard
 * Copyright (C) 2010 Antoine Mercier-Linteau
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "usb_keyboard.h"
#include <avr/pgmspace.h>
#include <avr/delay.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <string.h>
#include "base.h"
#include "timer2.h"
#include "HardwareSerial.h"

PROGMEM char usbHidReportDescriptor[39] = {
	    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
	    0x09, 0x06,                    // USAGE (Keyboard)
	    0xa1, 0x01,                    // COLLECTION (Application)
	    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
	    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
	    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
	    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
	    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
	    0x75, 0x01,                    //   REPORT_SIZE (1)
	    0x95, 0x08,                    //   REPORT_COUNT (8)
	    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
	    //0x95, 0x01,                    //   REPORT_COUNT (1)
	    //0x75, 0x08,                    //   REPORT_SIZE (8)
	    //0x81, 0x03,                    //   INPUT (Cnst,Var,Abs)
	    0x95, 0x03,                    //   REPORT_COUNT (3)
	    0x75, 0x08,                    //   REPORT_SIZE (8)
	    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
	    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
	    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
	    0x19, 0x00,                    //   USAGE_MINIMUM (Keyboard ErrorRollOver)
	    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
	    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
	    0xc0                           // END_COLLECTION
	};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */

USBKeyboard::USBKeyboard():
		idleRate(0),
		idleCounter(0),
		key_timeout(0),
		start(0),
		end(0),
		depress(false)
{
	instance = this;

	/* RESET status: all port bits are inputs without pull-up.
	 * That's the way we need D+ and D-. Therefore we don't need any
	 * additional hardware initialization.
	 */
	usbInit();
	cli();
	usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
	unsigned char i = 0;
	while(--i){             /* fake USB disconnect for > 250 ms */
		_delay_ms(1);
	}
	usbDeviceConnect();
	sei();
}

void USBKeyboard::build_report(void)
{
    if(start == end)
    {
    	outgoing_report.modifiers = 0;
    	outgoing_report.key0 = 0;
    	outgoing_report.key1 = 0;
    	outgoing_report.key2 = 0;
    }
    else
    {
    	outgoing_report = report_buffer[start];
		start = (start + 1) % REPORT_BUFFER_SIZE;
		if(start != end) //Check if there is another key in the buffer
		{
			key_timeout = Timer2::millis() + KEY_TIMEOUT_MS;
		}
    }
}


usbMsgLen_t usbFunctionSetup(uchar data[8])
{
	usbRequest_t *rq = (usbRequest_t*)data;

	usbMsgPtr = (unsigned char*)&instance->outgoing_report;
	if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
	{    /* class request type */
		if(rq->bRequest == USBRQ_HID_GET_REPORT)
		{  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
			/* we only have one report type, so don't look at wValue */
			//buildReport(keyPressed());
			return sizeof(instance->outgoing_report);
		}
		else if(rq->bRequest == USBRQ_HID_GET_IDLE)
		{
			usbMsgPtr = &instance->idleRate;
			return 1;
		}
		else if(rq->bRequest == USBRQ_HID_SET_IDLE)
		{
			instance->idleRate = rq->wValue.bytes[1];
		}
	}
	else
	{
		/* no vendor specific requests implemented */
	}
	return 0;
}


char USBKeyboard::take_report( char keys[])
{
	if(verbosity)
	{
		Serial.print("Received report ");
		Serial.print(keys[0], BIN);
		Serial.print(keys[1], DEC);
		Serial.print(keys[2], DEC);
		Serial.println(keys[3], DEC);
	}

	if((end + 1) % REPORT_BUFFER_SIZE != start)
	{
		if(verbosity)
		{
			Serial.println("Added report to queue");
		}
		report_buffer[end].modifiers = keys[0];
		report_buffer[end].key0 = keys[1];
		report_buffer[end].key1 = keys[2];
		report_buffer[end].key2 = keys[3];

		if(end == start) //Check is key is alone in the buffer
		{
			key_timeout = Timer2::millis() + KEY_TIMEOUT_MS;
		}

		end = (end + 1) % REPORT_BUFFER_SIZE;

		return 0;
	}

	return 1;
}

void USBKeyboard::run()
{
	usbPoll();

	if(usbInterruptIsReady())
	{
		if(start != end && !depress)
		{

			depress = true;
			build_report();
			if(verbosity)
			{
				Serial.print_P(PSTR("Sending report "));
				Serial.print(outgoing_report.modifiers, BIN);
				Serial.print(' ');
				Serial.print(outgoing_report.key0, DEC);
				Serial.print(' ');
				Serial.print(outgoing_report.key1, DEC);
				Serial.print(' ');
				Serial.println(outgoing_report.key2, DEC);
			}
			usbSetInterrupt((unsigned char*)&outgoing_report, sizeof(outgoing_report));
			idleCounter = Timer2::millis();
		}
		else
		{
			if(idleRate == 0 || idleCounter + idleRate * 4 < Timer2::millis())
			{

				outgoing_report.key0 = 0;
				outgoing_report.key1 = 0;
				outgoing_report.key2 = 0;

				usbSetInterrupt((unsigned char*)&outgoing_report, sizeof(outgoing_report));
				if(outgoing_report.modifiers)
				{
					outgoing_report.modifiers = 0;
				}
				else
				{
					depress = false;
				}
			}

		}
	}

	if(start != end) //Check if there is a shortcut in the buffer
	{
		if(key_timeout < Timer2::millis())
		{
			//The first shortcut has been there for too long, remove it
			start = (start + 1) % REPORT_BUFFER_SIZE;
			key_timeout = Timer2::millis() + KEY_TIMEOUT_MS; /*Updates the timeout counter
			for the next shortcut.*/
			if(verbosity)
			{
				Serial.println_P(PSTR("Shortcut timed out. Something is wrong with the USB!"));
			}
		}
	}


}

#define NUMBER_OF_USAGE_IDS 232

char usage_id_none[] PROGMEM = "none";
char usage_id_0[] PROGMEM = "Reserved"; //Used as not set for the keyMote
char usage_id_1[] PROGMEM = "ErrorRollOver";
char usage_id_2[] PROGMEM = "POSTFail";
char usage_id_3[] PROGMEM = "ErrorUndefined";
char usage_id_4[] PROGMEM = "aA";
char usage_id_5[] PROGMEM = "bB";
char usage_id_6[] PROGMEM = "cC";
char usage_id_7[] PROGMEM = "dD";
char usage_id_8[] PROGMEM = "eE";
char usage_id_9[] PROGMEM = "fF";
char usage_id_10[] PROGMEM = "gG";
char usage_id_11[] PROGMEM = "hH";
char usage_id_12[] PROGMEM = "iI";
char usage_id_13[] PROGMEM = "jJ";
char usage_id_14[] PROGMEM = "kK";
char usage_id_15[] PROGMEM = "lL";
char usage_id_16[] PROGMEM = "mM";
char usage_id_17[] PROGMEM = "nN";
char usage_id_18[] PROGMEM = "oO";
char usage_id_19[] PROGMEM = "pP";
char usage_id_20[] PROGMEM = "qQ";
char usage_id_21[] PROGMEM = "rR";
char usage_id_22[] PROGMEM = "sS";
char usage_id_23[] PROGMEM = "tT";
char usage_id_24[] PROGMEM = "uU";
char usage_id_25[] PROGMEM = "vV";
char usage_id_26[] PROGMEM = "wW";
char usage_id_27[] PROGMEM = "xX";
char usage_id_28[] PROGMEM = "yY";
char usage_id_29[] PROGMEM = "zZ";
char usage_id_30[] PROGMEM = "1!";
char usage_id_31[] PROGMEM = "2@";
char usage_id_32[] PROGMEM = "3#";
char usage_id_33[] PROGMEM = "4$";
char usage_id_34[] PROGMEM = "5%";
char usage_id_35[] PROGMEM = "6^";
char usage_id_36[] PROGMEM = "7&";
char usage_id_37[] PROGMEM = "8*";
char usage_id_38[] PROGMEM = "9(";
char usage_id_39[] PROGMEM = "0)";
char usage_id_40[] PROGMEM = "Return(Enter)";
char usage_id_41[] PROGMEM = "Escape";
char usage_id_42[] PROGMEM = "Backspace";
char usage_id_43[] PROGMEM = "Tab";
char usage_id_44[] PROGMEM = "Spacebar";
char usage_id_45[] PROGMEM = "-_";
char usage_id_46[] PROGMEM = "=+";
char usage_id_47[] PROGMEM = "[{";
char usage_id_48[] PROGMEM = "]}";
char usage_id_49[] PROGMEM = "\\|"; //The backslash need to be escaped
char usage_id_50[] PROGMEM = "NonUS#and~";
char usage_id_51[] PROGMEM = ";:";
char usage_id_52[] PROGMEM = "'\"";
char usage_id_53[] PROGMEM = "`~";
char usage_id_54[] PROGMEM = ",<";
char usage_id_55[] PROGMEM = ".>";
char usage_id_56[] PROGMEM = "/?";
char usage_id_57[] PROGMEM = "CapsLock";
char usage_id_58[] PROGMEM = "F1";
char usage_id_59[] PROGMEM = "F2";
char usage_id_60[] PROGMEM = "F3";
char usage_id_61[] PROGMEM = "F4";
char usage_id_62[] PROGMEM = "F5";
char usage_id_63[] PROGMEM = "F6";
char usage_id_64[] PROGMEM = "F7";
char usage_id_65[] PROGMEM = "F8";
char usage_id_66[] PROGMEM = "F9";
char usage_id_67[] PROGMEM = "F10";
char usage_id_68[] PROGMEM = "F11";
char usage_id_69[] PROGMEM = "F12";
char usage_id_70[] PROGMEM = "PrintScreen";
char usage_id_71[] PROGMEM = "ScrollLock";
char usage_id_72[] PROGMEM = "Pause";
char usage_id_73[] PROGMEM = "Insert";
char usage_id_74[] PROGMEM = "Home";
char usage_id_75[] PROGMEM = "PageUp";
char usage_id_76[] PROGMEM = "Delete";
char usage_id_77[] PROGMEM = "End";
char usage_id_78[] PROGMEM = "PageDown";
char usage_id_79[] PROGMEM = "RightArrow";
char usage_id_80[] PROGMEM = "LeftArrow";
char usage_id_81[] PROGMEM = "DownArrow";
char usage_id_82[] PROGMEM = "UpArrow";
char usage_id_83[] PROGMEM = "NumLock";
char usage_id_84[] PROGMEM = "Keypad/";
char usage_id_85[] PROGMEM = "Keypad*";
char usage_id_86[] PROGMEM = "Keypad-";
char usage_id_87[] PROGMEM = "Keypad+";
char usage_id_88[] PROGMEM = "KeypadEnter";
char usage_id_89[] PROGMEM = "Kepad1";
char usage_id_90[] PROGMEM = "Kepad2";
char usage_id_91[] PROGMEM = "Kepad3";
char usage_id_92[] PROGMEM = "Kepad4";
char usage_id_93[] PROGMEM = "Kepad5";
char usage_id_94[] PROGMEM = "Kepad6";
char usage_id_95[] PROGMEM = "Kepad7";
char usage_id_96[] PROGMEM = "Kepad8";
char usage_id_97[] PROGMEM = "Kepad9";
char usage_id_98[] PROGMEM = "Kepad0";
char usage_id_99[] PROGMEM = "Keypad.";
char usage_id_100[] PROGMEM = "NonUS\and|";
char usage_id_101[] PROGMEM = "Application";
//The previous are all the key usages on a standard IBM 101-key keyboard
char usage_id_102[] PROGMEM = "Power";
char usage_id_103[] PROGMEM = "Keypad=";
char usage_id_104[] PROGMEM = "F13";
char usage_id_105[] PROGMEM = "F14";
char usage_id_106[] PROGMEM = "F15";
char usage_id_107[] PROGMEM = "F16";
char usage_id_108[] PROGMEM = "F17";
char usage_id_109[] PROGMEM = "F18";
char usage_id_110[] PROGMEM = "F19";
char usage_id_111[] PROGMEM = "F20";
char usage_id_112[] PROGMEM = "F21";
char usage_id_113[] PROGMEM = "F22";
char usage_id_114[] PROGMEM = "F23";
char usage_id_115[] PROGMEM = "F24";
char usage_id_116[] PROGMEM = "Execute";
char usage_id_117[] PROGMEM = "Help";
char usage_id_118[] PROGMEM = "Menu";
char usage_id_119[] PROGMEM = "Select";
char usage_id_120[] PROGMEM = "Stop";
char usage_id_121[] PROGMEM = "Again";
char usage_id_122[] PROGMEM = "Undo";
char usage_id_123[] PROGMEM = "Cut";
char usage_id_124[] PROGMEM = "Copy";
char usage_id_125[] PROGMEM = "Paste";
char usage_id_126[] PROGMEM = "Find";
char usage_id_127[] PROGMEM = "Mute";
char usage_id_128[] PROGMEM = "VolumeUp";
char usage_id_129[] PROGMEM = "VolumeDown";
char usage_id_130[] PROGMEM = "LockingCapsLock";
char usage_id_131[] PROGMEM = "LockingNumLock";
char usage_id_132[] PROGMEM = "LockingScrollLock";
char usage_id_133[] PROGMEM = "KeypadComma";
char usage_id_134[] PROGMEM = "KeypadEqualSign";
char usage_id_135[] PROGMEM = "International1";
char usage_id_136[] PROGMEM = "International2";
char usage_id_137[] PROGMEM = "International3";
char usage_id_138[] PROGMEM = "International4";
char usage_id_139[] PROGMEM = "International5";
char usage_id_140[] PROGMEM = "International6";
char usage_id_141[] PROGMEM = "International7";
char usage_id_142[] PROGMEM = "International8";
char usage_id_143[] PROGMEM = "International9";
char usage_id_144[] PROGMEM = "LANG1";
char usage_id_145[] PROGMEM = "LANG2";
char usage_id_146[] PROGMEM = "LANG3";
char usage_id_147[] PROGMEM = "LANG4";
char usage_id_148[] PROGMEM = "LANG5";
char usage_id_149[] PROGMEM = "LANG6";
char usage_id_150[] PROGMEM = "LANG7";
char usage_id_151[] PROGMEM = "LANG8";
char usage_id_152[] PROGMEM = "LANG9";
char usage_id_153[] PROGMEM = "AlternateErase";
char usage_id_154[] PROGMEM = "SysReq/Attention";
char usage_id_155[] PROGMEM = "Cancel";
char usage_id_156[] PROGMEM = "Clear";
char usage_id_157[] PROGMEM = "Prior";
char usage_id_158[] PROGMEM = "Return";
char usage_id_159[] PROGMEM = "Separator";
char usage_id_160[] PROGMEM = "Out";
char usage_id_161[] PROGMEM = "Oper";
char usage_id_162[] PROGMEM = "Clear/Again";
char usage_id_163[] PROGMEM = "CrSel/Props";
char usage_id_164[] PROGMEM = "ExSel";
/* This block is reserved in the standard
 * char usage_id_165[] PROGMEM = "";
 * char usage_id_166[] PROGMEM = "";
 * char usage_id_167[] PROGMEM = "";
 * char usage_id_168[] PROGMEM = "";
 * char usage_id_169[] PROGMEM = "";
 * char usage_id_170[] PROGMEM = "";
 * char usage_id_171[] PROGMEM = "";
 * char usage_id_172[] PROGMEM = "";
 * char usage_id_173[] PROGMEM = "";
 * char usage_id_174[] PROGMEM = "";
 * char usage_id_175[] PROGMEM = "";*/
char usage_id_176[] PROGMEM = "Keypad00";
char usage_id_177[] PROGMEM = "Keypad000";
char usage_id_178[] PROGMEM = "ThousandsSeparator";
char usage_id_179[] PROGMEM = "DecimalSeparator";
char usage_id_180[] PROGMEM = "CurrencyUnit";
char usage_id_181[] PROGMEM = "CurrencySub-Unit";
char usage_id_182[] PROGMEM = "Keypad(";
char usage_id_183[] PROGMEM = "Keypad)";
char usage_id_184[] PROGMEM = "Keypad}";
char usage_id_185[] PROGMEM = "Keypad{";
char usage_id_186[] PROGMEM = "KeypadTab";
char usage_id_187[] PROGMEM = "KeypadBackspace";
char usage_id_188[] PROGMEM = "KeypadA";
char usage_id_189[] PROGMEM = "KeypadB";
char usage_id_190[] PROGMEM = "KeypadC";
char usage_id_191[] PROGMEM = "KeypadD";
char usage_id_192[] PROGMEM = "KeypadE";
char usage_id_193[] PROGMEM = "KeypadF";
char usage_id_194[] PROGMEM = "KeypadXOR";
char usage_id_195[] PROGMEM = "Keypad^";
char usage_id_196[] PROGMEM = "Keypad%";
char usage_id_197[] PROGMEM = "Keypad<";
char usage_id_198[] PROGMEM = "Keypad>";
char usage_id_199[] PROGMEM = "Keypad&";
char usage_id_200[] PROGMEM = "Keypad&&";
char usage_id_201[] PROGMEM = "Keypad|";
char usage_id_202[] PROGMEM = "Keypad||";
char usage_id_203[] PROGMEM = "Keypad:";
char usage_id_204[] PROGMEM = "Keypad#";
char usage_id_205[] PROGMEM = "KeypadSpace";
char usage_id_206[] PROGMEM = "Keypad@";
char usage_id_207[] PROGMEM = "Keypad!";
char usage_id_208[] PROGMEM = "KeypadMemoryStore";
char usage_id_209[] PROGMEM = "KeypadMemoryRecall";
char usage_id_210[] PROGMEM = "KeypadMemoryAdd";
char usage_id_211[] PROGMEM = "KeypadMemoryClear";
char usage_id_212[] PROGMEM = "KeypadMemorySubtract";
char usage_id_213[] PROGMEM = "KeypadMemoryMultiply";
char usage_id_214[] PROGMEM = "KeypadMemoryDivide";
char usage_id_215[] PROGMEM = "Keypad+/-";
char usage_id_216[] PROGMEM = "KeypadClear";
char usage_id_217[] PROGMEM = "KeypadClearEntry";
char usage_id_218[] PROGMEM = "KeypadBinary";
char usage_id_219[] PROGMEM = "KeypadOctal";
char usage_id_220[] PROGMEM = "KeypadDecimal";
char usage_id_221[] PROGMEM = "KeypadHexadecimal";
/* This block is reserved in the standard
 * char usage_id_222[] PROGMEM = "";
 * char usage_id_223[] PROGMEM = "";*/
char usage_id_224[] PROGMEM = "LeftControl";
char usage_id_225[] PROGMEM = "LeftShift";
char usage_id_226[] PROGMEM = "LeftAlt";
char usage_id_227[] PROGMEM = "LeftGUI";
char usage_id_228[] PROGMEM = "RightControl";
char usage_id_229[] PROGMEM = "RightShift";
char usage_id_230[] PROGMEM = "RightAlt";
char usage_id_231[] PROGMEM = "RightGUI";

PGM_P usage_id_table[] PROGMEM =
{
	usage_id_0,
	usage_id_1,
	usage_id_2,
	usage_id_3,
	usage_id_4,
	usage_id_5,
	usage_id_6,
	usage_id_7,
	usage_id_8,
	usage_id_9,
	usage_id_10,
	usage_id_11,
	usage_id_12,
	usage_id_13,
	usage_id_14,
	usage_id_15,
	usage_id_16,
	usage_id_17,
	usage_id_18,
	usage_id_19,
	usage_id_20,
	usage_id_21,
	usage_id_22,
	usage_id_23,
	usage_id_24,
	usage_id_25,
	usage_id_26,
	usage_id_27,
	usage_id_28,
	usage_id_29,
	usage_id_30,
	usage_id_31,
	usage_id_32,
	usage_id_33,
	usage_id_34,
	usage_id_35,
	usage_id_36,
	usage_id_37,
	usage_id_38,
	usage_id_39,
	usage_id_40,
	usage_id_41,
	usage_id_42,
	usage_id_43,
	usage_id_44,
	usage_id_45,
	usage_id_46,
	usage_id_47,
	usage_id_48,
	usage_id_49,
	usage_id_50,
	usage_id_51,
	usage_id_52,
	usage_id_53,
	usage_id_54,
	usage_id_55,
	usage_id_56,
	usage_id_57,
	usage_id_58,
	usage_id_59,
	usage_id_60,
	usage_id_61,
	usage_id_62,
	usage_id_63,
	usage_id_64,
	usage_id_65,
	usage_id_66,
	usage_id_67,
	usage_id_68,
	usage_id_69,
	usage_id_70,
	usage_id_71,
	usage_id_72,
	usage_id_73,
	usage_id_74,
	usage_id_75,
	usage_id_76,
	usage_id_77,
	usage_id_78,
	usage_id_79,
	usage_id_80,
	usage_id_81,
	usage_id_82,
	usage_id_83,
	usage_id_84,
	usage_id_85,
	usage_id_86,
	usage_id_87,
	usage_id_88,
	usage_id_89,
	usage_id_90,
	usage_id_91,
	usage_id_92,
	usage_id_93,
	usage_id_94,
	usage_id_95,
	usage_id_96,
	usage_id_97,
	usage_id_98,
	usage_id_99,
	usage_id_100,
	usage_id_101,
	usage_id_102,
	usage_id_103,
	usage_id_104,
	usage_id_105,
	usage_id_106,
	usage_id_107,
	usage_id_108,
	usage_id_109,
	usage_id_110,
	usage_id_111,
	usage_id_112,
	usage_id_113,
	usage_id_114,
	usage_id_115,
	usage_id_116,
	usage_id_117,
	usage_id_118,
	usage_id_119,
	usage_id_120,
	usage_id_121,
	usage_id_122,
	usage_id_123,
	usage_id_124,
	usage_id_125,
	usage_id_126,
	usage_id_127,
	usage_id_128,
	usage_id_129,
	usage_id_130,
	usage_id_131,
	usage_id_132,
	usage_id_133,
	usage_id_134,
	usage_id_135,
	usage_id_136,
	usage_id_137,
	usage_id_138,
	usage_id_139,
	usage_id_140,
	usage_id_141,
	usage_id_142,
	usage_id_143,
	usage_id_144,
	usage_id_145,
	usage_id_146,
	usage_id_147,
	usage_id_148,
	usage_id_149,
	usage_id_150,
	usage_id_151,
	usage_id_152,
	usage_id_153,
	usage_id_154,
	usage_id_155,
	usage_id_156,
	usage_id_157,
	usage_id_158,
	usage_id_159,
	usage_id_160,
	usage_id_161,
	usage_id_162,
	usage_id_163,
	usage_id_164,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_none,
	usage_id_176,
	usage_id_177,
	usage_id_178,
	usage_id_179,
	usage_id_180,
	usage_id_181,
	usage_id_182,
	usage_id_183,
	usage_id_184,
	usage_id_185,
	usage_id_186,
	usage_id_187,
	usage_id_188,
	usage_id_189,
	usage_id_190,
	usage_id_191,
	usage_id_192,
	usage_id_193,
	usage_id_194,
	usage_id_195,
	usage_id_196,
	usage_id_197,
	usage_id_198,
	usage_id_199,
	usage_id_200,
	usage_id_201,
	usage_id_202,
	usage_id_203,
	usage_id_204,
	usage_id_205,
	usage_id_206,
	usage_id_207,
	usage_id_208,
	usage_id_209,
	usage_id_210,
	usage_id_211,
	usage_id_212,
	usage_id_213,
	usage_id_214,
	usage_id_215,
	usage_id_216,
	usage_id_217,
	usage_id_218,
	usage_id_219,
	usage_id_220,
	usage_id_221,
	usage_id_none,
	usage_id_none,
	usage_id_224,
	usage_id_225,
	usage_id_226,
	usage_id_227,
	usage_id_228,
	usage_id_229,
	usage_id_230,
	usage_id_231
};

unsigned char USBKeyboard::string_to_usage_id( char* str )
{
	for(unsigned char i = 0; i< NUMBER_OF_USAGE_IDS; i++)
	{
		if(usage_id_table[i] == usage_id_none)
		{
			continue;
		}
		else if(!strncasecmp_P(str, (PGM_P)pgm_read_word(&(usage_id_table[i])), strlen(str)))
		{
			return i;
		}
	}

	return 0; //usage id matching string was not found.
}

char* USBKeyboard::usage_id_to_string( unsigned char id )
{
	if(id >= NUMBER_OF_USAGE_IDS)
	{
		char* str = (char*)malloc(strlen("Not mapped") + 1);
		strcpy(str, "Not mapped");
		return str;
	}
	PGM_P usage = (PGM_P)pgm_read_word(&(usage_id_table[id]));
	char* str = (char*)malloc(strlen_P(usage) + 1);
	if(str){ strcpy_P(str, usage); }
	return str;
}
