/* keypad.cpp - abstracts a keypad into a class
 * Copyright (C) 2010 Antoine Mercier-Linteau
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "keypad.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/delay.h>
#include "timer0.h"


Keypad::Keypad():
	keys_bouncing(false)
{

	//Set up keypad pins

	/*    PA1  PA7  PA3
	 *     |    |    |
	 * PA6--------------ROW1
	 *     |    |    |
	 * PA5--------------ROW2
	 *     |    |    |
	 * PA4--------------ROW3
	 *     |    |    |
	 * PA2--------------ROW4
	 *     |    |    |
	 *    COL1 COL2 COL3
	 */

	#define COL1 PIN5
	#define COL2 PIN6
	#define COL3 PIN7

	#define ROW1 PIN1
	#define ROW2 PIN2
	#define ROW3 PIN3
	#define ROW4 PIN4

	//Set colums as hi-z
	DDRA &= ~(_BV(COL1) + _BV(COL2) + _BV(COL3));
	PORTA &= ~(_BV(COL1) + _BV(COL2) + _BV(COL3));

	//Set rows as input
	DDRA &= ~(_BV(ROW1) + _BV(ROW2) + _BV(ROW3) + _BV(ROW4));

	GIMSK |= _BV(PCIE0); //Pin change interrupt 0 enabled

	PCMSK0 |= _BV(PCINT1) + _BV(PCINT2) + _BV(PCINT3) + _BV(PCINT4); /*Enable external interrupt
	on rows*/


	keys[0].state = UP + READ; keys[0].debounce = 0;
	keys[1].state = UP + READ; keys[1].debounce = 0;
	keys[2].state = UP + READ; keys[2].debounce = 0;
	keys[3].state = UP + READ; keys[3].debounce = 0;
	keys[4].state = UP + READ; keys[4].debounce = 0;
	keys[5].state = UP + READ; keys[5].debounce = 0;
	keys[6].state = UP + READ; keys[6].debounce = 0;
	keys[7].state = UP + READ; keys[7].debounce = 0;
	keys[8].state = UP + READ; keys[8].debounce = 0;
	keys[9].state = UP + READ; keys[9].debounce = 0;
	keys[10].state = UP + READ; keys[10].debounce = 0;
	keys[11].state = UP + READ; keys[11].debounce = 0;

}

void Keypad::scan(void)
{

	uint8_t keys_state[12];

	bool row1 = PINA & _BV(ROW1); //row1 is pressed
	bool row2 = PINA & _BV(ROW2); //row2 is pressed
	bool row3 = PINA & _BV(ROW3); //row3 is pressed
	bool row4 = PINA & _BV(ROW4); //row4 is pressed

	DDRA |= _BV(COL1); //Ground col 1
	if(row1 && !(PINA & _BV(ROW1))){ keys_state[0] = DOWN; } else { keys_state[0] = UP; }
	if(row2 && !(PINA & _BV(ROW2))){ keys_state[3] = DOWN; } else { keys_state[3] = UP; }
	if(row3 && !(PINA & _BV(ROW3))){ keys_state[6] = DOWN; } else { keys_state[6] = UP; }
	if(row4 && !(PINA & _BV(ROW4))){ keys_state[9] = DOWN; } else { keys_state[9] = UP; }
	DDRA &= ~(_BV(COL1)); //Z row 1


	DDRA |= _BV(COL2); //Ground col 2
	if(row1 && !(PINA & _BV(ROW1))){ keys_state[1] = DOWN; } else { keys_state[1] = UP; }
	if(row2 && !(PINA & _BV(ROW2))){ keys_state[4] = DOWN; } else { keys_state[4] = UP; }
	if(row3 && !(PINA & _BV(ROW3))){ keys_state[7] = DOWN; } else { keys_state[7] = UP; }
	if(row4 && !(PINA & _BV(ROW4))){ keys_state[10] = DOWN; } else { keys_state[10] = UP; }
	DDRA &= ~(_BV(COL2)); //Z col 2

	_delay_ms(1);

	DDRA |= _BV(COL3); //Ground col 3
	if(row1 && !(PINA & _BV(ROW1))){ keys_state[2] = DOWN; } else { keys_state[2] = UP; }
	if(row2 && !(PINA & _BV(ROW2))){ keys_state[5] = DOWN; } else { keys_state[5] = UP; }
	if(row3 && !(PINA & _BV(ROW3))){ keys_state[8] = DOWN; } else { keys_state[8] = UP; }
	if(row4 && !(PINA & _BV(ROW4))){ keys_state[11] = DOWN; } else { keys_state[11] = UP; }
	DDRA &= ~(_BV(COL3)); //Z col 3


	keys_up = 0;
	keys_down = 0;
	keys_bouncing = false;

	for(uint8_t i = 0; i < 12; i++)
	{
		if(keys[i].state & BOUNCING)
		{
			if(keys[i].debounce < Timer0::millis())
			{
				if(keys_state[i] != (keys[i].state & STATE))
				{
					keys[i].state = keys_state[i];
					if(keys_state[i] == DOWN) { keys_down++; }
					else{ keys_up++; }

				}
				keys[i].state &= ~BOUNCING; //Not needed...
			}
			else { keys_bouncing = true; }

		}
		else if(keys_state[i] != (keys[i].state & STATE))
		{
			keys[i].state |= BOUNCING;
			keys[i].debounce = Timer0::millis() + DEBOUNCE_MS;
			keys_bouncing = true;
		}

	}

}

unsigned char Keypad::available(void)
{
	return keys_up + keys_down ;
}

unsigned char Keypad::available_down(void)
{
	return keys_down;
}

uint8_t Keypad::read_down()
{
	for(uint8_t i = 0; i < 12; i++)
	{
		if(keys[i].state == 0) //!((keys[i].state & BOUNCING) || (keys[i].state & READ) || (keys[i].state & STATE)))
		{
			keys_down--;
			keys[i].state |= READ;
			return i + 1;
		}
	}

	return 0;
}

EMPTY_INTERRUPT(PCINT0_vect);


